﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SatikSudokuSolver
{
    public class SudokuBoard
    {
        byte[,] board = new byte[9, 9];
        private byte[] numbers = { 1, 2, 3, 4, 5, 6, 7, 8, 9 };

        private SudokuBoard()
        { }

        public SudokuBoard(string board, char empty = '_', char space = ' ')
        {
            ParseBoard(board, empty, space);
        }

        private void ParseBoard(string board, char empty = '_', char space = ' ')
        {
            int pos = 0;
            foreach (var c in board)
            {
                if (c == empty)
                    pos++;
                else
                    if (char.IsWhiteSpace(c))
                        continue;
                    else
                        if (char.IsDigit(c))
                        {
                            SetNumber(pos % 9, pos / 9, (byte)char.GetNumericValue(c));
                            pos++;
                        }
                        else
                            throw new FormatException("Input board is in wrong format.");
            }
        }

        public SudokuBoard Copy()
        {
            SudokuBoard copy = new SudokuBoard();
            copy.CopyFrom(this);
            return copy;
        }

        public void CopyFrom(SudokuBoard copyFrom)
        {
            this.board = copyFrom.board.Clone() as byte[,];
        }

        public void Print(char space = ' ', char empty = '_')
        {
            StringBuilder sb = new StringBuilder(200);

            for (int y = 0; y < 9; y++)
            {
                for (int x = 0; x < 9; x++)
                {
                    if (board[x, y] == 0)
                        sb.Append(empty);
                    else
                        sb.Append(board[x, y]);
                    sb.Append(space);
                }
                sb.Append(Environment.NewLine);
            }

            Console.WriteLine(sb.ToString());
        }

        public void SetNumber(int x, int y, byte number)
        {
            if (GetNumber(x, y) != 0)
                throw new ArgumentException("That field already contains a number.");

            board[x, y] = number;
        }

        public byte GetNumber(int x, int y)
        {
            return board[x, y];
        }

        public IEnumerable<byte> GetNumbersInColumn(byte column)
        {
            for (int i = 0; i < 9; i++)
            {
                yield return GetNumber(column, i);
            }
        }

        public IEnumerable<byte> GetNumbersInRow(byte column)
        {
            for (int i = 0; i < 9; i++)
            {
                yield return GetNumber(i, column);
            }
        }

        public IEnumerable<byte> GetNumbersInSquare(byte column, byte row)
        {
            byte baseX = (byte)((column / 3) * 3);
            byte baseY = (byte)((row / 3) * 3);

            for (int y = 0; y < 3; y++)
                for (int x = 0; x < 3; x++)
                {
                    yield return GetNumber(baseX + x, baseY + y);
                }
        }

        /// <summary>
        /// Returns lists of allowed numbers for all tiles - all numbers contained in same row, column and square are taken out
        /// </summary>
        public List<byte>[,] GetAllowedNumbers()
        {
            var allowedNumbers = new List<byte>[9, 9];

            for (byte y = 0; y < 9; y++)
            {
                for (byte x = 0; x < 9; x++)
                {
                    if (GetNumber(x, y) != 0) // this field is already filled in
                    {
                        allowedNumbers[x, y] = new List<byte>();
                    }
                    else
                    {
                        var forbiddenNumbers = GetNumbersInColumn(x).Union(GetNumbersInRow(y)).Union(GetNumbersInSquare(x, y));
                        allowedNumbers[x, y] = numbers.Except(forbiddenNumbers).ToList();
                    }
                }
            }

            return allowedNumbers;
        }

        /// <summary>
        /// Returns a value indicating whether the board is solved.
        /// </summary>
        public bool IsSolved
        {
            get
            {
                for (byte x = 0; x < 3; x++)
                    for (byte y = 0; y < 3; y++)
                    {
                        if (!VerifyWholeSequenceIsFilledIn(GetNumbersInSquare((byte)(x * 3), (byte)(y * 3))))
                            return false;
                    }

                for (byte i = 0; i < 9; i++)
                {
                    if (!VerifyWholeSequenceIsFilledIn(GetNumbersInRow(i)) ||
                      !VerifyWholeSequenceIsFilledIn(GetNumbersInColumn(i)))
                    {
                        return false;
                    }
                }

                return true;
            }
        }

        /// <summary>
        /// Verifies, that all numbers from 1 to 9 are contained in the enumerable.
        /// </summary>
        private bool VerifyWholeSequenceIsFilledIn(IEnumerable<byte> numbers)
        {
            for (byte i = 1; i < 10; i++)
            {
                if (!numbers.Contains(i))
                    return false;
            }

            return true;
        }


    }
}
