﻿namespace SudokuSolver
{
    using System.Collections.Generic;

    public sealed class Solver
    {
        private Field[,] fields;
        private List<int>[] rows;
        private List<int>[] columns;

        private bool numberAdded = false;
        private bool firstEvaluation = true;

        /// <summary>
        /// Initializes a new instance of the Solver class.
        /// </summary>
        /// <param name="fields"></param>
        public Solver(Field[,] fields)
        {
            this.fields = fields;
        }

        /// <summary>
        /// Main method that starts solving the sudoku
        /// </summary>
        public void Solve()
        {
            do
            {
                numberAdded = false;
                DefineRowsAndColumns();
                Evaluate();
                Scan();
            }
            while (numberAdded == true);

            AddRemaining(); // Pokud by algoritmus nedokazal soduku vyplnit, doplni to chybne cisla.
        }

        /// <summary>
        /// Defines content (numbers) of every row and column and save them into lists
        /// </summary>
        private void DefineRowsAndColumns()
        {
            rows = new List<int>[9];
            columns = new List<int>[9];

            for (int i = 0; i < 9; i++)
            {
                rows[i] = new List<int>();
                for (int j = 0; j < 9; j++)
                {
                    if (i == 0)
                        columns[j] = new List<int>();
                    if (fields[i, j].Number != 0)
                    {
                        rows[i].Add(fields[i, j].Number);
                        columns[j].Add(fields[i, j].Number);
                    }
                }
            }
        }

        /// <summary>
        /// Finds remaining empty fields and fills them with a possible number.
        /// Solves the problem with more solutions.
        /// </summary>
        private void AddRemaining()
        {
            for (int x = 0; x < 9; x++)
            {
                for (int y = 0; y < 9; y++)
                {
                    if (fields[x, y].Number == 0)
                    {
                        for (int k = 1; k <= 9; k++)
                        {
                            if (fields[x, y].Possibilities[k] == true &&
                                !rows[x].Contains(k) && 
                                !columns[y].Contains(k))
                            {
                                fields[x, y].Number = k;
                                rows[x].Add(k);
                                columns[y].Add(k);
                                break;
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Evaluates fields without number with possible numbers.
        /// </summary>
        private void Evaluate()
        {
            for (int i = 0; i < 9; i += 3)
            {
                for (int j = 0; j < 9; j += 3)
                {
                    EvaluateSquare(i, i + 3, j, j + 3);
                }
            }
            
            if (firstEvaluation == true)
                firstEvaluation = false;
        }

        /// <summary>
        /// Evaluates certain square of sudoku.
        /// </summary>
        /// <param name="startX"></param>
        /// <param name="endX"></param>
        /// <param name="startY"></param>
        /// <param name="endY"></param>
        private void EvaluateSquare(int startX, int endX, int startY, int endY)
        {
            List<int> squareCont = new List<int>();

            for (int i = startX; i < endX; i++)
            {
                for (int j = startY; j < endY; j++)
                {
                    if (fields[i, j].Number != 0)
                    {
                        squareCont.Add(fields[i, j].Number);
                    }
                }
            }

            for (int x = startX; x < endX; x++)
            {
                for (int y = startY; y < endY; y++)
                {
                    if (fields[x, y].Number == 0)
                    {
                        for (int k = 1; k <= 9; k++)
                        {
                            if (firstEvaluation == false)
                            {
                                if (fields[x, y].Possibilities[k] == true)
                                {
                                    if (squareCont.Contains(k) || rows[x].Contains(k) || columns[y].Contains(k))
                                    {
                                        fields[x, y].Possibilities[k] = false;
                                    }
                                }
                            }
                            else
                            {
                                if (!squareCont.Contains(k) && !rows[x].Contains(k) && !columns[y].Contains(k))
                                {
                                    fields[x, y].Possibilities[k] = true;
                                }
                            }
                        }
                    }
                    
                }
            }
        }

        /// <summary>
        /// Scanning method that searches for possible numbers which can add. 
        /// </summary>
        private void Scan()
        {
            for (int i = 1; i <= 9; i++)
            {
                Scan(i);
            }
            ScanSingleCandidate();
        }

        /// <summary>
        /// Scans squares, rows and columns with the specified number.
        /// </summary>
        /// <param name="number"></param>
        private void Scan(int number)
        {
            for (int i = 0; i < 9; i += 3)
            {
                for (int j = 0; j < 9; j += 3)
                {
                    ScanSquare(number, i, i + 3, j, j + 3);
                }
            }
            for (int i = 0; i < 9; i++)
            {
                ScanRow(number, i);
                ScanColumn(number, i);
            }
        }

        /// <summary>
        /// Finds single candidate and adds it in a field.
        /// </summary>
        private void ScanSingleCandidate()
        {
            // For every field 
            for (int x = 0; x < 9; x++)
            {
                for (int y = 0; y < 9; y++)
                {
                    if (fields[x, y].Number == 0)
                    {
                        int candidate = 0; 
                        bool single = false;

                        for (int i = 1; i <= 9; i++)
                        {
                            if (fields[x, y].Possibilities[i] == true)
                            {
                                if (candidate == 0)
                                {
                                    candidate = i;
                                    single = true;
                                }
                                else
                                {
                                    single = false;
                                    return;
                                }
                            }

                        }
                        if (single == true)
                        {
                            fields[x, y].Number = candidate;
                            numberAdded = true;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Scans certain square.
        /// </summary>
        /// <param name="number"></param>
        /// <param name="startX"></param>
        /// <param name="endX"></param>
        /// <param name="startY"></param>
        /// <param name="endY"></param>
        private void ScanSquare(int number, int startX, int endX, int startY, int endY)
        {
            int pos = 0; // Number of possibilities
            int x = -1;
            int y = -1;
            for (int i = startX; i < endX; i++)
            {
                for (int j = startY; j < endY; j++)
                {
                    if (fields[i, j].Number == number)
                        return;
                    if (fields[i, j].Number == 0 && fields[i, j].Possibilities[number] == true)
                    {
                        if (pos == 1)
                        {
                            return;
                        }
                        else
                        {
                            pos++;
                            x = i;
                            y = j;
                        }
                    }

                }
            }
            if (pos == 1)
            {
                fields[x, y].Number = number;
                numberAdded = true;
            }
        }

        /// <summary>
        /// Scans certain column.
        /// </summary>
        /// <param name="number"></param>
        /// <param name="columnIndex"></param>
        private void ScanColumn(int number, int columnIndex)
        {
            int pos = 0; // Number of possibilities
            int x = -1;
            for (int i = 0; i < 9; i++)
            {
                if (fields[i, columnIndex].Number == number)
                    return;
                if (fields[i, columnIndex].Number == 0 && fields[i, columnIndex].Possibilities[number] == true)
                {
                    if (pos == 1)
                    {
                        return;
                    }
                    else
                    {
                        pos++;
                        x = i;
                    }
                }
            }
            if (pos == 1)
            {
                fields[x, columnIndex].Number = number;
                numberAdded = true;
            }
        }

        /// <summary>
        /// Scans certain row.
        /// </summary>
        /// <param name="number"></param>
        /// <param name="rowIndex"></param>
        private void ScanRow(int number, int rowIndex)
        {
            int pos = 0; // Number of possibilities
            int y = -1;
            for (int i = 0; i < 9; i++)
            {
                if (fields[rowIndex, i].Number == number)
                    return;
                if (fields[rowIndex, i].Number == 0 && fields[rowIndex, i].Possibilities[number] == true)
                {
                    if (pos == 1)
                    {
                        return;
                    }
                    else
                    {
                        pos++;
                        y = i;
                    }
                }
            }
            if (pos == 1)
            {
                fields[rowIndex, y].Number = number;
                numberAdded = true;
            }
        }

    }
}
