package sudokusolver;

import java.util.HashSet;

/**
 * Třída slouží k vyřešení sudoku, pokud sudoku má nějaké řešení.
 * 
 */
public class ProcessSudoku {

    private final char[][] table;
        
    public ProcessSudoku(char[][] table) {
        this.table = table;     
    }
    
    public boolean solveSudoku() {
        this.fillWhatYouCan();
        return this.guessAndFillNumber();      
    }
    
    /**
     * Pokud jdou do sudoku nějaká čísla jednoznačně doplnit, tak je doplní
     */
    public void fillWhatYouCan() {
        boolean changed;
        
        do {
            changed = false;
            // Projde celé sudoku a do políček, u kterých lze jednoznačně vyplnit číslo, jej doplní.
            for(int i = 0; i < 9; i++) {
                for(int j = 0; j < 9; j++) {
                    if(table[i][j] == '_') {
                        char[] numbers = this.getPossibleNumbers(i, j);
                        if(numbers.length == 1) {
                            table[i][j] = numbers[0];
                            changed = true;
                        }
                    }
                }
            }
        }while(changed); // Pokud se v poslední iteraci doplnilo alespoň jedno políčko, tak opakuj
    }

    /**
     * Zjistí, která čísla jsou již vyplněna v daném řádku, sloupci a čtverci a vrátí ta čísla, která
     * ještě doplněna nejsou a je tedy možné je do daného políčka doplnit.
     * 
     * @param row       číslo řádku 
     * @param column    číslo sloupce
     * @return          všechna čísla, která je možno doplnit do políčka na souřadnicích (row, column)
     */
    private char[] getPossibleNumbers(int row, int column) {
        HashSet<Character> usedNumbers = new HashSet<>();
        
        for(int i = 0; i < 9; i++) {
            usedNumbers.add(table[row][i]);
            usedNumbers.add(table[i][column]);
            int y = ((row / 3) * 3) + (i / 3);
            int x = ((column / 3) * 3) + (i % 3);
            usedNumbers.add(table[y][x]);
        }
        
        usedNumbers.remove('_');
        char [] possibleNumbers = new char[(9 - usedNumbers.size())];
        int counter = 0;
            for(char i = '1'; i <= '9'; i++) {
                if(!usedNumbers.contains(i))
                    possibleNumbers[counter++] = i;
            }
        
        return possibleNumbers;
    }
    
    /**
     * Vytvoří dvourozměrné pole, kde jeden řádek odpovídá nevyplněnému políčku sudoku
     * a obsahuje čísla, která je možno do políčka doplnit
     * 
     * @return true pokud má sudoku řešení jinak false
     */
    private boolean guessAndFillNumber() {
        int unfillNumbersCount = this.getUnfillNumbersCount();
        char[][] possibleNumbers = new char[unfillNumbersCount][];
        int counter = 0;
        
        for(int i = 0; i < 9; i++) {
            for(int j = 0; j < 9; j++) {
                if(table[i][j] == '_')
                    possibleNumbers[counter++] = getPossibleNumbers(i, j);
            }
        }

        int[] position = new int[] {0,0};
        return this.depthFirstSearchAndFillNumbers(0, position , possibleNumbers);
    }
    
    /**
     * Rekurzivně doplňuje do jednotlivých políček sudoku možná čísla. Při každém zanoření
     * zkontroluje, zda doplněné číslo neporušuje pravidla sudoku (tedy je jediné pro daný řádek,
     * sloupec a čtverec). Pokud ano vynoří se o úroveň výše a zkusí doplnit další číslo.
     * 
     * @param order
     * @param position
     * @param possibleNumbers
     * @return true pokud má sudoku řešení jinak false
     */
    private boolean depthFirstSearchAndFillNumbers(int order,int[] position, char[][] possibleNumbers) {
        for(int i = 0; i < possibleNumbers[order].length; i++) {
            this.table[position[0]][position[1]] = possibleNumbers[order][i];
            if(checkFilledNumber(position[0], position[1]) && (isLastUnfilledNumber(position[0], position[1]) || depthFirstSearchAndFillNumbers((order+1),this.getNextUnfillNumberPosition(position[0], position[1]), possibleNumbers)))
                return true;
            else
                this.table[position[0]][position[1]] = '_';
        }
        return false;
    }
    
    /**
     * 
     * @param row   číslo řádku
     * @param column číslo sloupce
     * @return true pokud dané políčko je posledním nevyplněným políčkem v sudoku
     */
    private boolean isLastUnfilledNumber(int row, int column) {
        for(int i = row; i < 9; i++) {
            for(int j = 0; j < 9; j++) {
                if(table[i][j] == '_' && (i > row || j > column))
                    return false;
            }
        }
        return true;
    }
    
    /**
     * 
     * @param row       číslo řádku
     * @param column    číslo sloupce
     * @return          true pokud číslo v daném řádku a sloupci vyhovuje pravidlům sudoku
     */
    private boolean checkFilledNumber(int row, int column) {
        return checkRow(row) & checkColumn(column) & checkBox(row, column);
    }
    
    private boolean checkRow(int row) {        
        HashSet<Character> usedNumbers = new HashSet<>();
        
        for(int i = 0; i < 9; i++) {
            if(usedNumbers.contains(table[row][i]) && (table[row][i] != '_'))
                return false;
            else
                usedNumbers.add(table[row][i]);
        }
        return true;
    }
    
    private boolean checkColumn(int column) {        
        HashSet<Character> usedNumbers = new HashSet<>();
        
        for(int i = 0; i < 9; i++) {
            if(usedNumbers.contains(table[i][column]) && (table[i][column] != '_'))
                return false;
            else
                usedNumbers.add(table[i][column]);
        }
        
        return true;
    }
    
        private boolean checkBox(int row, int column) {        
        HashSet<Character> usedNumbers = new HashSet<>();
        
        for(int i = 0; i < 9; i++) {
            int y = ((row / 3) * 3) + (i / 3);
            int x = ((column / 3) * 3) + (i % 3);
            
            if(usedNumbers.contains(table[y][x]) && (table[y][x] != '_'))
                return false;
            else
                usedNumbers.add(table[y][x]);
        }
        
        return true;
    }
    
        /**
         * 
         * @return počet nevyplněných políček sudoku
         */
        private int getUnfillNumbersCount() {
        int counter = 0;
        for(int i = 0; i < 9; i++) {
            for(int j = 0; j < 9; j++) {
                if(table[i][j] == '_')
                    counter++;
            }
        }
        
        return counter;
    }
    
        /**
         * 
         * @param row       číslo řádku
         * @param column    číslo sloupce
         * @return          vrátí pozici následujícího nevyplněného políčka od políčka na
         *                  pozici (řádek, sloupec), tabulka se prochází zhora dolů, zleva doprava
         */
        private int[] getNextUnfillNumberPosition(int row, int column) {
        int[] position = new int[2];
        for(int i = row; i < 9; i++) {
            for(int j = 0; j < 9; j++) {
                if(table[i][j] == '_' && (i > row || j > column)) {
                    position[0] = i;
                    position[1] = j;
                    return position;
                }
                    
            }
        }
        return position;
    }
    
    public char[][] getTable() {    
        return this.table;
    }
}
